import random
import torch
import torch.backends.cudnn as cudnn
import numpy as np

cudnn.enabled = False
cudnn.benchmark = True

from .path import *


def count_trainable_params(model):
    count = 0
    for param in model.parameters():
        if param.requires_grad:
            count += param.numel()
    return count


def seed_torch(seed=3407):
    # random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


def loadWeight(path, net):
    print(f"count_trainable_params = {count_trainable_params(model=net)}")

    if checkFile(path):
        print(f"Loading net from : {path}")
        check = torch.load(path)
        net.load_state_dict(check)
    else:
        print("There is no weight to be loaded!!")


class EarlyStopping(object):
    def __init__(self, criterion: float = 0.0, patience: int = 7):
        """
        连续 patience 轮 criterion 指标不增加
        """
        self.EarlyStopping = False
        self.patience = patience
        self.count = 0
        self.criterion = criterion

    def CheckStopping(self, new_criterion):
        flag = False
        if new_criterion < self.criterion:
            self.count += 1
            print(f"EarlyStopping counter: {self.count} out of {self.patience}")
        else:
            print(f"EarlyStopping criterion : {self.criterion} => {new_criterion}")
            self.criterion = new_criterion
            self.count = 0
            flag = True
        if self.count >= self.patience:
            self.EarlyStopping = True
        return flag



from ..backbone.swin4skysense import sswin_l
from torch import nn
import torch.nn.functional as F


class BaseNet(nn.Module):
    def __init__(self, backbone):
        super(BaseNet, self).__init__()
        backbone_zoo = {"sswin_l": sswin_l}
        self.backbone = backbone_zoo[backbone](pretrained=False)

    def base_forward(self, x):
        h, w = x.shape[-2:]

        x = self.backbone.base_forward(x)[-1]
        x = self.head(x)
        x = F.interpolate(x, (h, w), mode="bilinear", align_corners=True)

        return x

    def forward(self, x, tta=False):
        if not tta:
            return self.base_forward(x)

        else:
            h, w = x.shape[-2:]
            scales = [0.5, 0.75, 1.0, 1.5, 2.0]

            final_result = None

            for scale in scales:
                cur_h, cur_w = int(h * scale), int(w * scale)
                cur_x = F.interpolate(x, size=(cur_h, cur_w), mode='bilinear', align_corners=True)

                out = F.softmax(self.base_forward(cur_x), dim=1)
                out = F.interpolate(out, (h, w), mode='bilinear', align_corners=True)
                final_result = out if final_result is None else (final_result + out)

                out = F.softmax(self.base_forward(cur_x.flip(3)), dim=1).flip(3)
                out = F.interpolate(out, (h, w), mode='bilinear', align_corners=True)
                final_result += out

            return final_result



import torch


class settings(object):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # loss only for mask2former
    class_weight = 2.0
    mask_weight = 5.0
    dice_weight = 5.0
    train_num_points = 12544
    deep_supervision = True
    no_object_weight = 0.1
    dec_layers = 10
    oversample_ratio = 3.0
    importance_sample_ratio = 0.75

    data_root = r"/scratch/liwanchun/Large_Size_Datasets"
    # data_root = r"weight"
    batch_size = 4
    num_workers = 2
    learn_rate = 0.0001

    def __init__(self,
                 data_type="FBP",
                 downsample=8,
                 backbone="swin_b",
                 segHead="mask2former",
                 isContext=False):
        self.isContext = isContext
        # data
        self.data_type = data_type
        self.downsample = downsample

        # model
        self.backbone = backbone
        self.segHead = segHead
        if data_type == "FBP":
            self.nclass = 25
        elif data_type == "HPD":
            self.nclass = 10
        else:
            self.nclass = 2
        self.ignore_index = -100 if self.nclass == 2 else 0

        self.resume = False
        # self.pth_resume = "weight/mask2former/FBP_d1_best.pth"
        self.pth_resume = f"weight/{segHead}/{data_type}_d{downsample}_1024_best.pth"
        # self.pth_name = f"{data_type}_d{downsample}_1024_fed"
        # self.pth_name = f"{data_type}_d{downsample}"
        self.pth_path = f"weight/{segHead}"

        # train
        self.max_epochs = 50 * downsample
        if data_type == "FBP":
            self.early_stopping_patience = max(2 * downsample + 5, 7)
            self.lr_patience = max(2 * downsample + 3, 5)
        else:
            self.early_stopping_patience = 3
            self.lr_patience = 5
        # self.early_stopping_patience = 7
        # self.lr_patience = 5
        print("########################################")
        print(f"max_epochs = {self.max_epochs}")
        print(f"early_stopping_patience = {self.early_stopping_patience}")
        print(f"lr_patience = {self.lr_patience}")
        print(f"isContext = {isContext}")
        print(f"feature encoder = {self.backbone}")
        print(f"segment decoder = {self.segHead}")
        print("########################################")



from .base import BaseNet
import torch
from torch import nn
import torch.nn.functional as F
import torchvision
from .lib_upernet.nn import SynchronizedBatchNorm2d  # , PrRoIPool2D
from .GCM import Global_Context_Module


def conv3x3(in_planes, out_planes, stride=1, has_bias=False):
    # 3x3 convolution with padding
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=has_bias)


def conv3x3_bn_relu(in_planes, out_planes, stride=1):
    return nn.Sequential(
        conv3x3(in_planes, out_planes, stride),
        SynchronizedBatchNorm2d(out_planes),
        nn.ReLU(inplace=True),
    )


class UPerNet(BaseNet):

    def __init__(self, backbone, nclass,
                 pool_scales=(1, 2, 3, 6),
                 fpn_inplanes=(256, 512, 1024, 2048), fc_dim=4096,
                 fpn_dim: int = 512, isContext: bool = False):
        super(UPerNet, self).__init__(backbone)
        fc_dim = self.backbone.channels[-1]
        fpn_inplanes = self.backbone.channels
        # print(self.backbone.channels)
        self.nclass = nclass
        # Global_Context_Module
        # TODO
        self.isContext = isContext
        print(f"isContext = {self.isContext}")
        if self.isContext:
            self.GCM = Global_Context_Module(channels=self.backbone.channels)

        # PPM Module
        self.ppm_pooling = []
        self.ppm_conv = []

        for scale in pool_scales:
            # we use the feature map size instead of input image size, so down_scale = 1.0
            # self.ppm_pooling.append(PrRoIPool2D(scale, scale, 1.))  # TODO
            self.ppm_pooling.append(nn.AdaptiveAvgPool2d(scale))

            self.ppm_conv.append(nn.Sequential(
                nn.Conv2d(fc_dim, 512, kernel_size=1, bias=False),
                SynchronizedBatchNorm2d(512),
                nn.ReLU(inplace=True)
            ))
        self.ppm_pooling = nn.ModuleList(self.ppm_pooling)
        self.ppm_conv = nn.ModuleList(self.ppm_conv)
        self.ppm_last_conv = conv3x3_bn_relu(fc_dim + len(pool_scales) * 512, fpn_dim, 1)

        # FPN Module
        self.fpn_in = []
        for fpn_inplane in fpn_inplanes[:-1]:  # skip the top layer
            self.fpn_in.append(nn.Sequential(
                nn.Conv2d(fpn_inplane, fpn_dim, kernel_size=1, bias=False),
                SynchronizedBatchNorm2d(fpn_dim),
                nn.ReLU(inplace=True)
            ))
        self.fpn_in = nn.ModuleList(self.fpn_in)

        self.fpn_out = []
        for i in range(len(fpn_inplanes) - 1):  # skip the top layer
            self.fpn_out.append(nn.Sequential(
                conv3x3_bn_relu(fpn_dim, fpn_dim, 1),
            ))
        self.fpn_out = nn.ModuleList(self.fpn_out)

        self.conv_fusion = conv3x3_bn_relu(len(fpn_inplanes) * fpn_dim, fpn_dim, 1)

        # input: Fusion out, input_dim: fpn_dim
        self.object_head = nn.Sequential(
            conv3x3_bn_relu(fpn_dim, fpn_dim, 1),
            nn.Conv2d(fpn_dim, self.nclass, kernel_size=1, bias=True)
        )

    def base_forward0(self, x):
        seg_size = x.shape[-2:]
        conv_out = self.backbone.base_forward(x)
        # for idx, f in enumerate(conv_out):
        #     print(idx, f.shape)
        conv5 = conv_out[-1]
        input_size = conv5.size()
        # TODO
        # roi = []  # fake rois, just used for pooling
        # for i in range(input_size[0]):  # batch size
        #     roi.append(torch.Tensor([i, 0, 0, input_size[3], input_size[2]]).view(1, -1))  # b, x0, y0, x1, y1
        # roi = torch.cat(roi, dim=0).type_as(conv5)
        ppm_out = [conv5]
        for pool_scale, pool_conv in zip(self.ppm_pooling, self.ppm_conv):
            # print(pool_conv)
            # print(pool_scale(conv5).shape)
            ppm_out.append(pool_conv(F.interpolate(
                # pool_scale(conv5, roi.detach()), #TODO
                pool_scale(conv5),
                (input_size[2], input_size[3]),
                mode='bilinear', align_corners=False)))
        # for idx, f in enumerate(ppm_out):
        #     print(idx, f.shape)
        ppm_out = torch.cat(ppm_out, 1)
        f = self.ppm_last_conv(ppm_out)

        fpn_feature_list = [f]
        for i in reversed(range(len(conv_out) - 1)):
            conv_x = conv_out[i]
            conv_x = self.fpn_in[i](conv_x)  # lateral branch

            f = F.interpolate(
                f, size=conv_x.size()[2:], mode='bilinear', align_corners=False)  # top-down branch
            f = conv_x + f

            fpn_feature_list.append(self.fpn_out[i](f))
        fpn_feature_list.reverse()  # [P2 - P5]

        output_size = fpn_feature_list[0].size()[2:]
        fusion_list = [fpn_feature_list[0]]
        for i in range(1, len(fpn_feature_list)):
            fusion_list.append(F.interpolate(
                fpn_feature_list[i],
                output_size,
                mode='bilinear', align_corners=False))
        fusion_out = torch.cat(fusion_list, 1)
        x = self.conv_fusion(fusion_out)
        out = self.object_head(x)
        # print("out::", out.shape)
        return F.interpolate(out, size=seg_size, mode='bilinear', align_corners=False)

    def base_forward_context(self, xs):
        x1, x2, x3 = xs
        with torch.no_grad():
            f2 = self.backbone.base_forward(x2)
            f3 = self.backbone.base_forward(x3)
        f1 = self.backbone.base_forward(x1)

        seg_size = x1.shape[-2:]
        conv_out = self.GCM(f1, f2, f3)

        conv5 = conv_out[-1]
        input_size = conv5.size()
        ppm_out = [conv5]
        for pool_scale, pool_conv in zip(self.ppm_pooling, self.ppm_conv):
            ppm_out.append(pool_conv(F.interpolate(
                # pool_scale(conv5, roi.detach()), #TODO
                pool_scale(conv5),
                (input_size[2], input_size[3]),
                mode='bilinear', align_corners=False)))

        ppm_out = torch.cat(ppm_out, 1)
        f = self.ppm_last_conv(ppm_out)

        fpn_feature_list = [f]
        for i in reversed(range(len(conv_out) - 1)):
            conv_x = conv_out[i]
            conv_x = self.fpn_in[i](conv_x)  # lateral branch

            f = F.interpolate(
                f, size=conv_x.size()[2:], mode='bilinear', align_corners=False)  # top-down branch
            f = conv_x + f

            fpn_feature_list.append(self.fpn_out[i](f))
        fpn_feature_list.reverse()  # [P2 - P5]

        output_size = fpn_feature_list[0].size()[2:]
        fusion_list = [fpn_feature_list[0]]
        for i in range(1, len(fpn_feature_list)):
            fusion_list.append(F.interpolate(
                fpn_feature_list[i],
                output_size,
                mode='bilinear', align_corners=False))
        fusion_out = torch.cat(fusion_list, 1)
        x = self.conv_fusion(fusion_out)
        out = self.object_head(x)
        return F.interpolate(out, size=seg_size, mode='bilinear', align_corners=False)

    def base_forward(self, x):
        if self.isContext:
            return self.base_forward_context(x)
        else:
            return self.base_forward0(x)



import os
import cv2
import random
from PIL import Image
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data.dataset import Dataset
import random
import h5py


def flip_image(image, label):
    index = random.choice([0, 1, -1])
    return cv2.flip(image, index), cv2.flip(label, index)


class Single_View(Dataset):
    def __init__(self, root_path, isSingleClass=False):
        super(Single_View, self).__init__()

        self.image_path = os.path.join(root_path, "img")
        self.label_path = os.path.join(root_path, "label")
        self.data_list = []
        self.isSingleClass = isSingleClass

        for image_name in os.listdir(self.image_path):
            self.data_list.append(image_name[:-4])

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx):
        name = self.data_list[idx]
        img = cv2.imread(os.path.join(self.image_path, name + '.jpg')).astype("uint8")
        ann = np.array(Image.open(os.path.join(self.label_path, name + '.png'))).astype("uint8")
        if self.isSingleClass:
            ann[ann < 100] = 0
            ann[ann >= 100] = 1
        maskTensor = torch.from_numpy(ann).type(torch.LongTensor)
        out = {
            "img": torch.from_numpy(np.transpose(img / 255, (2, 0, 1))).type(torch.FloatTensor),
            "mask": maskTensor
        }
        return out


class Single_View_HDF5(Dataset):
    def __init__(self, HDF5path):
        super(Single_View_HDF5, self).__init__()
        with h5py.File(HDF5path, "r") as hf:
            self.images = hf["images"][:]
            self.labels = hf["labels"][:]
        hf.close()

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        img = self.images[idx]
        ann = self.labels[idx]
        maskTensor = torch.from_numpy(ann).type(torch.LongTensor)
        out = {
            "img": torch.from_numpy(np.transpose(img / 255, (2, 0, 1))).type(torch.FloatTensor),
            "mask": maskTensor
        }
        return out



# -*- coding: utf-8 -*-
# File   : test_prroi_pooling2d.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 18/02/2018
#
# This file is part of Jacinle.

import unittest

import torch
import torch.nn as nn
import torch.nn.functional as F

from jactorch.utils.unittest import TorchTestCase

from prroi_pool import PrRoIPool2D


class TestPrRoIPool2D(TorchTestCase):
    def test_forward(self):
        pool = PrRoIPool2D(7, 7, spatial_scale=0.5)
        features = torch.rand((4, 16, 24, 32)).cuda()
        rois = torch.tensor([
            [0, 0, 0, 14, 14],
            [1, 14, 14, 28, 28],
        ]).float().cuda()

        out = pool(features, rois)
        out_gold = F.avg_pool2d(features, kernel_size=2, stride=1)

        self.assertTensorClose(out, torch.stack((
            out_gold[0, :, :7, :7],
            out_gold[1, :, 7:14, 7:14],
        ), dim=0))

    def test_backward_shapeonly(self):
        pool = PrRoIPool2D(2, 2, spatial_scale=0.5)

        features = torch.rand((4, 2, 24, 32)).cuda()
        rois = torch.tensor([
            [0, 0, 0, 4, 4],
            [1, 14, 14, 18, 18],
        ]).float().cuda()
        features.requires_grad = rois.requires_grad = True
        out = pool(features, rois)

        loss = out.sum()
        loss.backward()

        self.assertTupleEqual(features.size(), features.grad.size())
        self.assertTupleEqual(rois.size(), rois.grad.size())


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : functional.py
# Author : Jiayuan Mao, Tete Xiao
# Email  : maojiayuan@gmail.com, jasonhsiao97@gmail.com
# Date   : 07/13/2018
#
# This file is part of PreciseRoIPooling.
# Distributed under terms of the MIT license.
# Copyright (c) 2017 Megvii Technology Limited.

import torch
import torch.autograd as ag

try:
    from os.path import join as pjoin, dirname
    from torch.utils.cpp_extension import load as load_extension
    root_dir = pjoin(dirname(__file__), 'src')
    _prroi_pooling = load_extension(
        '_prroi_pooling',
        [pjoin(root_dir, 'prroi_pooling_gpu.c'), pjoin(root_dir, 'prroi_pooling_gpu_impl.cu')],
        verbose=True
    )
except ImportError:
    raise ImportError('Can not compile Precise RoI Pooling library.')

__all__ = ['prroi_pool2d']


class PrRoIPool2DFunction(ag.Function):
    @staticmethod
    def forward(ctx, features, rois, pooled_height, pooled_width, spatial_scale):
        assert 'FloatTensor' in features.type() and 'FloatTensor' in rois.type(), \
                'Precise RoI Pooling only takes float input, got {} for features and {} for rois.'.format(features.type(), rois.type())

        pooled_height = int(pooled_height)
        pooled_width = int(pooled_width)
        spatial_scale = float(spatial_scale)

        features = features.contiguous()
        rois = rois.contiguous()
        params = (pooled_height, pooled_width, spatial_scale)

        if features.is_cuda:
            output = _prroi_pooling.prroi_pooling_forward_cuda(features, rois, *params)
            ctx.params = params
            # everything here is contiguous.
            ctx.save_for_backward(features, rois, output)
        else:
            raise NotImplementedError('Precise RoI Pooling only supports GPU (cuda) implememtations.')

        return output

    @staticmethod
    def backward(ctx, grad_output):
        features, rois, output = ctx.saved_tensors
        grad_input = grad_coor = None

        if features.requires_grad:
            grad_output = grad_output.contiguous()
            grad_input = _prroi_pooling.prroi_pooling_backward_cuda(features, rois, output, grad_output, *ctx.params)
        if rois.requires_grad:
            grad_output = grad_output.contiguous()
            grad_coor = _prroi_pooling.prroi_pooling_coor_backward_cuda(features, rois, output, grad_output, *ctx.params)

        return grad_input, grad_coor, None, None, None


prroi_pool2d = PrRoIPool2DFunction.apply




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# Author : Jiayuan Mao, Tete Xiao
# Email  : maojiayuan@gmail.com, jasonhsiao97@gmail.com
# Date   : 07/13/2018
# 
# This file is part of PreciseRoIPooling.
# Distributed under terms of the MIT license.
# Copyright (c) 2017 Megvii Technology Limited.

from .prroi_pool import *




import os
import torch
import torch.optim as optim
import torch.nn as nn
from typing import Any, Dict, List, Set
import copy
import itertools
from .model.semseg.upernet import UPerNet

# from dataset import build_dataloader
from .configs.args import settings
from .utils.net_utils import loadWeight, seed_torch, EarlyStopping
from .utils.metrics import Evaluator


def build_model(args: settings):
    seed_torch()
    print(f"feature encoder : {args.backbone}")
    print(f"segment decoder : {args.segHead}")
    if args.segHead == "upernet":
        model = UPerNet(backbone=args.backbone, nclass=args.nclass, isContext=args.isContext)
    else:
        raise ValueError(f"Invalid seg head: {args.segHead}")
    if args.resume:
        loadWeight(path=args.pth_resume, net=model)
    # model.to(args.device)
    return model


def build_data(args: settings):
    print(f"data type : {args.data_type}")
    print(f"downsample : {args.downsample}")
    trainDataLoader = build_dataloader(
        data_path=os.path.join(args.data_root, f"{args.data_type}_train_d{args.downsample}_1024.h5"),
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        isSingleClass=args.data_type == "FBP")
    valDataLoader = build_dataloader(
        data_path=os.path.join(args.data_root, f"{args.data_type}_val_d{args.downsample}_1024.h5"),
        batch_size=1,  # val batch size = 1
        isSingleClass=args.data_type == "FBP")

    return trainDataLoader, valDataLoader


def build_optimizer_mask2former(model, lr):
    weight_decay_norm = 0.0  # cfg.SOLVER.WEIGHT_DECAY_NORM
    weight_decay_embed = 0.0  # cfg.SOLVER.WEIGHT_DECAY_EMBED

    defaults = {}
    BASE_LR = lr
    defaults["lr"] = BASE_LR  # cfg.SOLVER.BASE_LR
    defaults["weight_decay"] = 0.05  # cfg.SOLVER.WEIGHT_DECAY

    norm_module_types = (
        torch.nn.BatchNorm1d,
        torch.nn.BatchNorm2d,
        torch.nn.BatchNorm3d,
        torch.nn.SyncBatchNorm,
        # NaiveSyncBatchNorm inherits from BatchNorm2d
        torch.nn.GroupNorm,
        torch.nn.InstanceNorm1d,
        torch.nn.InstanceNorm2d,
        torch.nn.InstanceNorm3d,
        torch.nn.LayerNorm,
        torch.nn.LocalResponseNorm,
    )

    params: List[Dict[str, Any]] = []
    memo: Set[torch.nn.parameter.Parameter] = set()
    for module_name, module in model.named_modules():
        for module_param_name, value in module.named_parameters(recurse=False):
            if not value.requires_grad:
                continue
            # Avoid duplicating parameters
            if value in memo:
                continue
            memo.add(value)

            hyperparams = copy.copy(defaults)
            if "backbone" in module_name:
                hyperparams["lr"] = hyperparams["lr"] * 0.1  # cfg.SOLVER.BACKBONE_MULTIPLIER
            if (
                    "relative_position_bias_table" in module_param_name
                    or "absolute_pos_embed" in module_param_name
            ):
                print(module_param_name)
                hyperparams["weight_decay"] = 0.0
            if isinstance(module, norm_module_types):
                hyperparams["weight_decay"] = weight_decay_norm
            if isinstance(module, torch.nn.Embedding):
                hyperparams["weight_decay"] = weight_decay_embed
            params.append({"params": [value], **hyperparams})

    def maybe_add_full_model_gradient_clipping(optim):
        # detectron2 doesn't have full model gradient clipping now
        clip_norm_val = 0.1  # cfg.SOLVER.CLIP_GRADIENTS.CLIP_VALUE
        enable = (
            # cfg.SOLVER.CLIP_GRADIENTS.ENABLED
            # and cfg.SOLVER.CLIP_GRADIENTS.CLIP_TYPE == "full_model"
                True
                and clip_norm_val > 0.0
        )

        class FullModelGradientClippingOptimizer(optim):
            def step(self, closure=None):
                all_params = itertools.chain(*[x["params"] for x in self.param_groups])
                torch.nn.utils.clip_grad_norm_(all_params, clip_norm_val)
                super().step(closure=closure)

        return FullModelGradientClippingOptimizer if enable else optim

    optimizer_type = "ADAMW"  # cfg.SOLVER.OPTIMIZER
    if optimizer_type == "SGD":
        optimizer = maybe_add_full_model_gradient_clipping(torch.optim.SGD)(
            params, BASE_LR, momentum=cfg.SOLVER.MOMENTUM
        )
    elif optimizer_type == "ADAMW":
        optimizer = maybe_add_full_model_gradient_clipping(torch.optim.AdamW)(
            params, BASE_LR
        )
    else:
        raise NotImplementedError(f"no optimizer type {optimizer_type}")
    # if not cfg.SOLVER.CLIP_GRADIENTS.CLIP_TYPE == "full_model":
    #     optimizer = maybe_add_gradient_clipping(cfg, optimizer)
    return optimizer

def build_evaluator(args: settings, now_criterion: float = 0):
    evaluator = Evaluator(args.nclass, args.ignore_index)
    earlystopping = EarlyStopping(criterion=now_criterion, patience=args.early_stopping_patience)
    return evaluator, earlystopping




import torch
from torch import nn
import torch.nn.functional as F


# pytorch 1.7.0
class Global_Context_Module(nn.Module):
    def __init__(self, channels, nHeads: int = 2, nLayers: int = 3):
        super(Global_Context_Module, self).__init__()
        assert len(channels) == 4  # 四组特征的通道数
        assert nLayers == 3
        c1, c2, c3, c4 = channels

        self.attention12 = nn.ModuleList()
        self.attention13 = nn.ModuleList()
        self.Squeue = nn.ModuleList()
        for c in (c2, c3, c4):
            self.attention12.append(nn.MultiheadAttention(embed_dim=c, num_heads=nHeads))
            self.attention13.append(nn.MultiheadAttention(embed_dim=c, num_heads=nHeads))
            self.Squeue.append(nn.Conv2d(c * nLayers, c, kernel_size=1, stride=1, padding=0))

    @staticmethod
    def att(q, kv, attention, saveName=None):
        # q,kv [bs,c,h,w]
        # attention nn.MultiheadAttention
        # (L,N,E)
        # where L is the target sequence length, N is the batch size, E is the embedding dimension.
        # return q
        # 将图像特征展平为序列长度
        bs, c, h, w = q.shape
        q = q.flatten(2).permute(2, 0, 1)  # h*w,bs,c
        kv = kv.flatten(2).permute(2, 0, 1)

        output, _ = attention(q, kv, kv)
        output = output.permute(1, 2, 0)  # (batch_size, embed_dim, seq_len)
        output = output.view(bs, c, h, w)  # (batch_size, embed_dim, height, width)

        return output

    def forward(self, f1, f2, f3):
        """
        :param f1: 来自小视场图像的特征，有四组 已经有梯度
        :param f2: 来自中视场图像的特征，有四组
        :param f3: 来自大视场图像的特征，有四组
        :return: 融合之后的特征，四组
        """
        outFeats = []

        for i, x in enumerate(zip(f1, f2, f3)):
            fx1, fx2, fx3 = x
            if i == 0:
                outFeats.append(fx1)
                continue
            fx2.requires_grad = True
            fx3.requires_grad = True
            Fx = torch.cat((fx1,
                            self.att(q=fx1, kv=fx2, attention=self.attention12[i - 1]),
                            self.att(q=fx1, kv=fx3, attention=self.attention13[i - 1])),
                           dim=1)
            outFeats.append(self.Squeue[i - 1](Fx))
        return outFeats



from distutils.core import setup
from setuptools import find_packages

with open("README.rst", "r") as f:
  long_description = f.read()

setup(name='skysenseremote',  # 包名
        version='1.1.3',  # 版本号
      description='A skysense model package',
      long_description=long_description,
      author='mikezhou_talk',
      author_email='bodang@whu.edu.cn',
      license='Apache license 2.0',
      packages=find_packages(),
      platforms=["all"],
      classifiers=[
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Natural Language :: Chinese (Simplified)',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries'
      ],
      install_requires = ["numpy", "pillow"]
      )



import os
import random
from typing import Sequence
import numpy as np
import mmcv
import torch
from torchvision import transforms
import cv2

fbp_map = {
    0: [200, 0, 0],
    1: [0, 200, 0],
    2: [150, 250, 0],
    3: [150, 200, 150],
    4: [200, 0, 200],
    5: [150, 0, 250],
    6: [150, 150, 250],
    7: [200, 150, 200],
    8: [250, 200, 0],
    9: [200, 200, 0],
    10: [0, 0, 200],
    11: [250, 0, 150],
    12: [0, 150, 200],
    13: [0, 200, 250],
    14: [150, 200, 250],
    15: [250, 250, 250],
    16: [200, 200, 200],
    17: [200, 150, 150],
    18: [250, 200, 150],
    19: [150, 150, 0],
    20: [250, 150, 150],
    21: [250, 150, 0],
    22: [250, 200, 250],
    23: [200, 150, 0]
}

def normalize_function(img):  # 更快
    # Mean values used to pre-training the pre-trained backbone models
    mean = [123.675, 116.28, 103.53]
    std = [58.395, 57.12, 57.375]
    mean = np.array(mean, dtype=np.float32)
    std = np.array(std, dtype=np.float32)
    results = mmcv.imnormalize(img, mean, std, to_rgb=True)
    return torch.from_numpy(results.transpose((2, 0, 1))).float()


def photo_metric_distortion(img: np.ndarray,
                            brightness_delta: int = 32,
                            contrast_range: Sequence[float] = (0.5, 1.5),
                            saturation_range: Sequence[float] = (0.5, 1.5),
                            hue_delta: int = 18) -> np.ndarray:
    def convert(img: np.ndarray, alpha: int = 1, beta: int = 0) -> np.ndarray:
        img = img.astype(np.float32) * alpha + beta
        img = np.clip(img, 0, 255)
        return img.astype(np.uint8)

    def brightness(img: np.ndarray) -> np.ndarray:
        if random.randint(0, 1):
            return convert(img, beta=random.uniform(-brightness_delta, brightness_delta))
        return img

    def contrast(img: np.ndarray) -> np.ndarray:
        if random.randint(0, 1):
            return convert(img, alpha=random.uniform(contrast_range[0], contrast_range[1]))
        return img

    def saturation(img: np.ndarray) -> np.ndarray:
        if random.randint(0, 1):
            img_hsv = mmcv.bgr2hsv(img)
            img_hsv[:, :, 1] = convert(img_hsv[:, :, 1], alpha=random.uniform(saturation_range[0], saturation_range[1]))
            img = mmcv.hsv2bgr(img_hsv)
        return img

    def hue(img: np.ndarray) -> np.ndarray:
        if random.randint(0, 1):
            img_hsv = mmcv.bgr2hsv(img)
            img_hsv[:, :, 0] = (img_hsv[:, :, 0].astype(int) + random.randint(-hue_delta, hue_delta)) % 180
            img = mmcv.hsv2bgr(img_hsv)
        return img

    # random brightness
    img = brightness(img)

    # mode == 0 --> do random contrast first
    # mode == 1 --> do random contrast last
    mode = random.randint(0, 1)
    if mode == 1:
        img = contrast(img)

    # random saturation
    img = saturation(img)

    # random hue
    img = hue(img)

    # random contrast
    if mode == 0:
        img = contrast(img)

    return img


def GridMask(image, grid_size=16, cell_size=32,
             mask_ratio=0.25):
    grid_size = image.shape[0] // cell_size
    mask = np.zeros((grid_size, grid_size))
    mask[1::2, 1::2] = 1
    for i in range(grid_size):
        for j in range(grid_size):
            if mask[i, j] == 1:
                image[i * cell_size: (i + 1) * cell_size, j * cell_size: (j + 1) * cell_size, :] = \
                    [123.675, 116.28, 103.53]
    return image


def BlockMask(image, grid_size=16, cell_size=32,
              mask_ratio=0.25):
    grid_size = image.shape[0] // cell_size
    mask = np.zeros((grid_size, grid_size))
    r = random.randint(0, grid_size // 2)
    c = random.randint(0, grid_size // 2)
    mask[r:r + grid_size // 2, c:c + grid_size // 2] = 1
    for i in range(grid_size):
        for j in range(grid_size):
            if mask[i, j] == 1:
                image[i * cell_size: (i + 1) * cell_size, j * cell_size: (j + 1) * cell_size, :] = \
                    [123.675, 116.28, 103.53]
    return image


def RandomMask(image, grid_size=16, cell_size=32,
               mask_ratio=0.25):
    # grid_size = image.shape[0] // cell_size
    mask = np.random.choice([0, 1], size=(grid_size, grid_size), p=[1 - mask_ratio, mask_ratio])
    for i in range(grid_size):
        for j in range(grid_size):
            if mask[i, j] == 1:
                image[i * cell_size: (i + 1) * cell_size, j * cell_size: (j + 1) * cell_size, :] = \
                    [123.675, 116.28, 103.53]
    return image


def checkPath(path):
    if not os.path.exists(path):
        os.makedirs(path)


def checkFile(path):
    return os.path.isfile(path)


# un use

class Normalize(object):
    # rgb
    # imagenet_mean = [0.485, 0.456, 0.406]
    # imagenet_std = [0.229, 0.224, 0.225]
    def __init__(self, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
        self.mean = mean
        self.std = std

    def __call__(self, img):
        # img:opencv:bgr
        # cv2.cvtColor(img, cv2.COLOR_BGR2RGB, img)
        img = np.array(img[:, :, ::-1]).astype(np.float32)
        img /= 255.0
        img -= self.mean
        img /= self.std

        return img


class ToTensor(object):
    def __call__(self, img):
        # swap color axis because
        # numpy image: H x W x C
        # torch image: C X H X W
        img = np.array(img).astype(np.float32).transpose((2, 0, 1))
        img = torch.from_numpy(img).float()

        return img


# 将 Normalize 和 ToTensor 组合成一个函数
def transform_function(img):
    composed_transforms = transforms.Compose([
        Normalize(),
        ToTensor(),
    ])

    return composed_transforms(img)


def color_distortion(image):
    # 随机改变图像的亮度、对比度和饱和度
    alpha = 1.0 + np.random.uniform(-0.5, 0.5)  # 亮度
    beta = 0.5 + np.random.uniform(-0.5, 0.5)  # 对比度
    saturation_scale = 0.5 + np.random.uniform(0, 1)  # 饱和度

    image = cv2.convertScaleAbs(image, alpha=alpha, beta=beta)
    hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    # 调整饱和度
    hsv[:, :, 1] = np.clip(hsv[:, :, 1] * saturation_scale, 0, 255)

    distorted_image = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
    return distorted_image


def gaussian_noise(image):
    row, col, ch = image.shape
    mean = 0
    var = 0.1
    sigma = var ** 0.5
    gauss = np.random.normal(mean, sigma, (row, col, ch))
    noisy = np.clip(image + gauss * 255, 0, 255).astype(np.uint8)
    return noisy
